@@ -1,5 +1,3 @@
-from __future__ import print_function
-
 import fnmatch
 import os
 import re
@@ -7,12 +5,14 @@ import signal
 import sys
 import threading
 import time
-from glob import glob
 
-from .compat import queue
+from .compat import queue, glob
 from .ipc import ProcessGroup
-from .logger import DefaultLogger
-from .utils import resolve_spec, is_watchdog_supported, is_watchman_supported
+from .logger import DefaultLogger, SilentLogger
+from .utils import default
+from .utils import is_watchdog_supported
+from .utils import is_watchman_supported
+from .utils import resolve_spec
 from .worker import Worker, is_active, get_reloader
 
 
@@ -38,7 +38,7 @@ class FileMonitorProxy(object):
         # if the glob does not match any files then go ahead and pass
         # the pattern to the monitor anyway incase it is just a file that
         # is currently missing
-        for p in glob(path) or [path]:
+        for p in glob(path, recursive=True) or [path]:
             if not any(x.match(p) for x in self.ignore_files):
                 self.monitor.add_path(p)
 
@@ -69,6 +69,13 @@ class FileMonitorProxy(object):
         self.change_event.set()
 
 
+class WorkerResult:
+    BROKEN_PIPE = 'broken_pipe'
+    EXIT = 'exit'
+    FILE_CHANGED = 'file_changed'
+    RELOAD_REQUEST = 'reload_request'
+
+
 class Reloader(object):
     """
     A wrapper class around a file monitor which will handle changes by
@@ -82,6 +89,7 @@ class Reloader(object):
         monitor_factory,
         logger,
         reload_interval=1,
+        shutdown_interval=1,
         worker_args=None,
         worker_kwargs=None,
         ignore_files=None,
@@ -92,9 +100,9 @@ class Reloader(object):
         self.ignore_files = ignore_files
         self.monitor_factory = monitor_factory
         self.reload_interval = reload_interval
+        self.shutdown_interval = shutdown_interval
         self.logger = logger
         self.monitor = None
-        self.worker = None
         self.group = ProcessGroup()
 
     def run(self):
@@ -108,9 +116,11 @@ class Reloader(object):
         self._start_monitor()
         try:
             while True:
-                if not self._run_worker():
-                    self._wait_for_changes()
-                time.sleep(self.reload_interval)
+                result = self._run_worker()
+                if result == WorkerResult.EXIT:
+                    result = self._wait_for_changes()
+                if result != WorkerResult.RELOAD_REQUEST:
+                    time.sleep(self.reload_interval)
         except KeyboardInterrupt:
             pass
         finally:
@@ -136,81 +146,29 @@ class Reloader(object):
             self._restore_signals()
 
     def _run_worker(self):
-        self.worker = Worker(
+        worker = Worker(
             self.worker_path, args=self.worker_args, kwargs=self.worker_kwargs
         )
-        self.worker.start()
-
-        try:
-            # register the worker with the process group
-            self.group.add_child(self.worker.pid)
-
-            self.logger.info("Starting monitor for PID %s." % self.worker.pid)
-            self.monitor.clear_changes()
-
-            while not self.monitor.is_changed() and self.worker.is_alive():
-                try:
-                    cmd = self.worker.pipe.recv(timeout=self.reload_interval)
-                except queue.Empty:
-                    continue
-
-                if cmd is None:
-                    if self.worker.is_alive():
-                        # the worker socket has died but the process is still
-                        # alive (somehow) so wait a brief period to see if it
-                        # dies on its own - if it does die then we want to
-                        # treat it as a crash and wait for changes before
-                        # reloading, if it doesn't die then we want to force
-                        # reload the app immediately because it probably
-                        # didn't die due to some file changes
-                        time.sleep(self.reload_interval)
-                    break
-
-                if cmd[0] == 'reload':
-                    break
-
-                if cmd[0] == 'watch':
-                    for path in cmd[1]:
-                        self.monitor.add_path(path)
-
-                else:  # pragma: no cover
-                    raise RuntimeError('received unknown command')
-
-        except KeyboardInterrupt:
-            if self.worker.is_alive():
-                self.logger.info('Waiting for server to exit ...')
-                time.sleep(self.reload_interval)
-            raise
-
-        finally:
-            if self.worker.is_alive():
-                self.logger.info(
-                    'Killing server with PID %s.' % self.worker.pid
-                )
-                self.worker.terminate()
-                self.worker.join()
-
-            else:
-                self.worker.join()
-                self.logger.info(
-                    'Server with PID %s exited with code %d.'
-                    % (self.worker.pid, self.worker.exitcode)
-                )
-
-        self.monitor.clear_changes()
-
-        force_restart = self.worker.terminated
-        self.worker = None
-        return force_restart
+        return _run_worker(
+            worker,
+            self.monitor,
+            self.group,
+            self.logger,
+            self.reload_interval,
+            self.shutdown_interval,
+        )
 
     def _wait_for_changes(self):
-        self.logger.info('Waiting for changes before reloading.')
-        while not self.monitor.wait_for_change(
-            self.reload_interval
-        ):  # pragma: nocover
-            pass
-
-        self.monitor.clear_changes()
+        self.logger.info('Press ENTER or change a file to reload.')
+        worker = Worker(__name__ + '.wait_main')
+        return _run_worker(
+            worker,
+            self.monitor,
+            self.group,
+            SilentLogger(),
+            self.reload_interval,
+            self.shutdown_interval,
+        )
 
     def _start_monitor(self):
         proxy = FileMonitorProxy(self.logger, self.ignore_files)
@@ -241,6 +199,95 @@ class Reloader(object):
             signal.signal(signal.SIGHUP, signal.SIG_DFL)
 
 
+def _run_worker(
+    worker, monitor, process_group, logger, reload_interval, shutdown_interval
+):
+    worker.start()
+    result = WorkerResult.EXIT
+
+    try:
+        # register the worker with the process group
+        process_group.add_child(worker.pid)
+
+        logger.info('Starting monitor for PID %s.' % worker.pid)
+        monitor.clear_changes()
+
+        while worker.is_alive():
+            if monitor.is_changed():
+                result = WorkerResult.FILE_CHANGED
+                break
+
+            try:
+                cmd = worker.pipe.recv(timeout=0.5)
+            except queue.Empty:
+                continue
+
+            if cmd is None:
+                if worker.is_alive():
+                    # the worker socket has died but the process is still
+                    # alive (somehow) so wait a brief period to see if it
+                    # dies on its own - if it does die then we want to
+                    # treat it as a crash and wait for changes before
+                    # reloading, if it doesn't die then we want to force
+                    # reload the app immediately because it probably
+                    # didn't die due to some file changes
+                    time.sleep(reload_interval)
+
+                if worker.is_alive():
+                    logger.info('Broken pipe to server, triggering a reload.')
+                    result = WorkerResult.BROKEN_PIPE
+
+                else:
+                    logger.debug('Broken pipe to server, looks like a crash.')
+                break
+
+            if cmd[0] == 'reload':
+                logger.debug('Server triggered a reload.')
+                result = WorkerResult.RELOAD_REQUEST
+                break
+
+            if cmd[0] == 'watch':
+                for path in cmd[1]:
+                    monitor.add_path(path)
+
+            else:  # pragma: no cover
+                raise RuntimeError('received unknown command')
+
+        if worker.is_alive() and shutdown_interval is not None:
+            logger.info('Gracefully killing the server.')
+            worker.kill(soft=True)
+            worker.wait(shutdown_interval)
+
+    except KeyboardInterrupt:
+        if worker.is_alive():
+            logger.info('Received interrupt, waiting for server to exit ...')
+            if shutdown_interval is not None:
+                worker.wait(shutdown_interval)
+        raise
+
+    finally:
+        if worker.is_alive():
+            logger.info('Server did not exit, forcefully killing.')
+            worker.kill()
+            worker.join()
+
+        else:
+            worker.join()
+        logger.debug('Server exited with code %d.' % worker.exitcode)
+
+    monitor.clear_changes()
+    return result
+
+
+def wait_main():
+    try:
+        reloader = get_reloader()
+        input('')
+        reloader.trigger_reload()
+    except KeyboardInterrupt:
+        pass
+
+
 def find_default_monitor_factory(logger):
     spec = os.getenv('HUPPER_DEFAULT_MONITOR')
     if spec:
@@ -269,7 +316,9 @@ def find_default_monitor_factory(logger):
 def start_reloader(
     worker_path,
     reload_interval=1,
+    shutdown_interval=default,
     verbose=1,
+    logger=None,
     monitor_factory=None,
     worker_args=None,
     worker_kwargs=None,
@@ -290,11 +339,18 @@ def start_reloader(
     is invoking ``start_reloader`` in the first place.
 
     ``reload_interval`` is a value in seconds and will be used to throttle
-    restarts.
+    restarts. Default is ``1``.
+
+    ``shutdown_interval`` is a value in seconds and will be used to trigger
+    a graceful shutdown of the server. Set to ``None`` to disable the graceful
+    shutdown. Default is the same as ``reload_interval``.
 
     ``verbose`` controls the output. Set to ``0`` to turn off any logging
     of activity and turn up to ``2`` for extra output. Default is ``1``.
 
+    ``logger``, if supplied, supersedes ``verbose`` and should be an object
+    implementing :class:`hupper.interfaces.ILogger`.
+
     ``monitor_factory`` is an instance of
     :class:`hupper.interfaces.IFileMonitorFactory`. If left unspecified, this
     will try to create a :class:`hupper.watchdog.WatchdogFileMonitor` if
@@ -313,16 +369,21 @@ def start_reloader(
     if is_active():
         return get_reloader()
 
-    logger = DefaultLogger(verbose)
+    if logger is None:
+        logger = DefaultLogger(verbose)
 
     if monitor_factory is None:
         monitor_factory = find_default_monitor_factory(logger)
 
+    if shutdown_interval is default:
+        shutdown_interval = reload_interval
+
     reloader = Reloader(
         worker_path=worker_path,
         worker_args=worker_args,
         worker_kwargs=worker_kwargs,
         reload_interval=reload_interval,
+        shutdown_interval=shutdown_interval,
         monitor_factory=monitor_factory,
         logger=logger,
         ignore_files=ignore_files,
