@@ -134,7 +134,6 @@ class Worker(object):
         self.worker_args = args
         self.worker_kwargs = kwargs
         self.pipe, self._child_pipe = ipc.Pipe()
-        self.terminated = False
         self.pid = None
         self.process = None
         self.exitcode = None
@@ -150,7 +149,7 @@ class Worker(object):
             pipe=self._child_pipe,
         )
         self.process = ipc.spawn(
-            'hupper.worker.worker_main',
+            __name__ + '.worker_main',
             kwargs=kw,
             pass_fds=[self._child_pipe.r_fd, self._child_pipe.w_fd],
         )
@@ -167,15 +166,17 @@ class Worker(object):
         if self.exitcode is not None:
             return False
         if self.process:
-            return self.process.poll() is None
+            return ipc.wait(self.process, timeout=0) is None
         return False
 
-    def terminate(self):
-        self.terminated = True
-        self.process.terminate()
+    def kill(self, soft=False):
+        return ipc.kill(self.process, soft=soft)
+
+    def wait(self, timeout=None):
+        return ipc.wait(self.process, timeout=timeout)
 
     def join(self):
-        self.exitcode = self.process.wait()
+        self.exitcode = self.wait()
 
         if self.stdin_termios:
             ipc.restore_termios(sys.stdin.fileno(), self.stdin_termios)
